<#
.SYNOPSIS
    SSH & Telnet automation of jBASE jShell sessions.
    Automatic login and execution on a remote jBASE/T24 server of specified
    non-blocking jShell tasks and commands.
    SFTP/FTP upload of jBASE Basic source files.

.DESCRIPTION
    This script allows to automate jBASE jShell sessions via SSH connections
    on Unix/Linux and via Telnet connections on Windows.
    Tasks supported are:
    - Compilation and execution of jBC subroutines.
    - Evaluation of jBC expressions / statements.
    - Execution of jShell non-interactive / non-blocking commands.
    - Uploading of jBC source files.

.PARAMETER Upload
    Flag requesting upload via SFTP/FTP of a local file to a remote jBASE/T24 server.
.PARAMETER SrcFileToUpload
    Full name of file to upload.
.PARAMETER Compile
    Flag requesting compilation of a jBASE jBC source file on a remote jBASE/T24 server.
.PARAMETER SubRoutineToCompile
    Name of jBC subroutine to compile.
.PARAMETER Run
    Flag requesting the execution of a jBASE jBC subroutine on a remote jBASE/T24 server.
.PARAMETER SubRoutineToRun
    Name of jBC subroutine to call / execute.
.PARAMETER EvalFile
    Flag requesting source file creation, compilation and execution of jBC
    expressions / statements on a remote jBASE/T24 server.
.PARAMETER SrcFilePath
    Full name of local file containing the jBC source code sent for evaluation.
.PARAMETER EvalString
    Flag requesting source file creation, compilation and execution of jBC
    expressions / statements on a remote jBASE/T24 server.
.PARAMETER SrcBuffer
    Base64 encoded string buffer containing the jBC source code sent for evaluation.
.PARAMETER Jsh
    Flag requesting execution of a non-interactive / non-blocking jBASE jShell command
    on a remote jBASE/T24 server.
.PARAMETER jshCommand
    String containing a non-interactive / non-blocking jBASE jShell command.
.PARAMETER UseSSH
    This option indicates that an SSH connection will be used to communicate with the
    remote jBASE/T24 server. If the option is not included then a Telnet connection
    will be used instead.
.PARAMETER Dummyparm
    Used when the script is called from notepad++ nppexec plugin

.EXAMPLE
    jBASE-Actions.ps1 -Upload "E:\Development\HELLO.ROUTINE" -SSH
    Description
    -----------
    Invokes uploading of local jBC source filename to remote jBASE/T24 server.
    Creates an SSH connection with the jBASE/T24 server to upload the file.
.EXAMPLE
    jBASE-Actions.ps1 -Compile HELLO.ROUTINE -Verbose
    Description
    -----------
    Invokes compilation and cataloguing of jBC subroutine "HELLO.ROUTINE"
    on the remote jBASE/T24 server. The -Verbose option outputs additional
    information showing all the steps involved in the execution of the task.
.EXAMPLE
    jBASE-Actions.ps1 -Run HELLO.ROUTINE Hello World!
    Description
    -----------
    Calls jBC subroutine "HELLO.ROUTINE" for execution with two parameters.
.EXAMPLE
    jBASE-Actions.ps1 -EvalFile "E:\Development\jBCEvalCode.txt"
    Description
    -----------
    Reads contents at the specified filepath, encode to base64 and send it
    to remote jBASE/T24 server for evaluation.
.EXAMPLE
    jBASE-Actions.ps1 -EvalString "Q1JUICJIZWxsbyEi"
    Description
    -----------
    Reads the base64 encoded string containing jBC source code and send it to
    remote jBASE/T24 server for evaluation.
    After decoded, the contents of the string are: CRT "Hello!"
.EXAMPLE
    jBASE-Actions.ps1 -Jsh "jdiag" -SSH
    Description
    -----------
    Returns the output obtained after executing the non-interactive jShell command.
    Creates an SSH connection with the jBASE/T24 server to run the command.

.NOTES
    Name:     jBASE-Actions
    Version:  1.1.0
    Author:   trainer@ambdev.net
    Created:  December 2012
    LastEdit: 2014/03/10

    Requires Ketul Patel's updated version of SharpSSH library, already included
    in the installation package. Original SharpSSH Library by Tamir Gal.

    This script makes use of code found in "SharpSsh Functions" by Joel Bennett.

.LINK
    http://ketulpatel.wordpress.com/2010/05/13/enhanced-sharpssh
.LINK
    http://www.tamirgal.com/blog/page/SharpSSH.aspx
.LINK
    http://poshcode.org/1010
.LINK
    http://www.youtube.com/user/moboan
.LINK
    https://bitbucket.org/moboan/jbcide
#>

[CmdletBinding()]

Param (

    [parameter(
        mandatory = $true,
        ParameterSetName = "Upload",
        Position = 0
    )]
    [switch] $Upload,

    [parameter(
        ParameterSetName = "Upload",
        Position = 1
    )]
    [ValidateNotNullOrEmpty()]
    [string] $SrcFileToUpload = "",

    [parameter(
        mandatory = $true,
        ParameterSetName = "Compile",
        Position = 0
    )]
    [switch] $Compile,

    [parameter(
        ParameterSetName = "Compile",
        Position = 1
    )]
    [ValidateNotNullOrEmpty()]
    [string] $SubRoutineToCompile = "",

    [parameter(
        mandatory = $true,
        ParameterSetName = "Run",
        Position = 0
    )]
    [switch] $Run,

    [parameter(
        ParameterSetName = "Run",
        Position = 1
    )]
    [ValidateNotNullOrEmpty()]
    [string] $SubRoutineToRun = "",

    [parameter(
        ParameterSetName = "Run",
        Position = 2,
        ValueFromRemainingArguments = $true
    )]
    [string[]] $SubRoutineArgs ,

    [parameter(
        mandatory = $true,
        ParameterSetName = "EvalFile",
        Position = 0
    )]
    [switch] $EvalFile,

    [parameter(
        ParameterSetName = "EvalFile",
        Position = 1
    )]
    [ValidateNotNullOrEmpty()]
    [string] $SrcFilePath = "",

    [parameter(
        mandatory = $true,
        ParameterSetName = "EvalString",
        Position = 0
    )]
    [switch] $EvalString,

    [parameter(
        ParameterSetName = "EvalString",
        Position = 1
    )]
    [ValidateNotNullOrEmpty()]
    [String] $SrcBuffer = "",

    [parameter(
        mandatory = $true,
        ParameterSetName = "Jsh",
        Position = 0
    )]
    [switch] $Jsh,

    [parameter(
        ParameterSetName = "Jsh",
        Position = 1
    )]
    [ValidateNotNullOrEmpty()]
    [string] $jshCommand = "",

    [alias("SSH")]
    [switch] $UseSSH,

    [parameter(
        ValueFromPipeline = $true
    )]
    [string] $Dummyparm

)


$t = [diagnostics.stopwatch]::startnew()

# load the config file
$scriptDir = Split-Path -Path $MyInvocation.MyCommand.Definition -Parent
$ConfigFilePath = Join-Path -Path $scriptDir -ChildPath "jbcConfig.ps1"
$config = (. $ConfigFilePath)

$ModulePath = Join-Path -Path $scriptDir -ChildPath $config.ModuleName
Import-Module -Name $ModulePath | Out-Null

switch ($PsCmdlet.ParameterSetName) {

    "Upload"     { Start-Upload $SrcFileToUpload; break }
    "Compile"    { Invoke-JbcCompiler $SubRoutineToCompile; break }
    "Run"        { Start-SubRoutine $SubRoutineToRun $SubRoutineArgs; break }
    "EvalFile"   { Start-JbcSourceFileEval $SrcFilePath; break }
    "EvalString" { Start-JbcSourceStringEval $SrcBuffer; break }
    "Jsh"        { Invoke-JshCommand $jshCommand; break }

}

$t.stop()
Write-Verbose ("`nElapsed time: {0}" -f $t.Elapsed.TotalSeconds)
