<#
    jBASE Unix / Linux SSH Actions Module

#>

# load the SharpSSH namespace
[void][reflection.assembly]::LoadFrom( (Join-Path $scriptDir $config.AssemblyPath) )


# create SSH session using SharpSSH library
function New-SshSession {
    Param (
        [string] $HostName,
        [int] $PortNum,
        [string] $UserName,
        [string] $UserPwd
    )

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [New-SshSession] - Connecting via SSH to jBASE server on $($HostName):$($PortNum)"

    $SshShell = new-object Tamir.SharpSsh.SshShell $HostName, $UserName, $UserPwd
    $SshShell.RemoveTerminalEmulationCharacters = $true
    # pattern should match otherwise stream read operations will wait forever
    $SshShell.ExpectPattern = $config.Prompt + "\e\[K\e\[J$|jBASE debugger->$|\*\* Error "

    Try {
        $SshShell.Connect($PortNum)
        Write-Verbose "$(Get-Date -f $config.Dtfmt) [New-SshSession] - SSH session created!"
        return $SshShell
    } Catch [System.Exception] {
        Write-Warning "Verfy connection properties in 'jbcConfig'"
        Write-Host $_.Exception.GetType().FullName
        Write-Host $_.Exception.Message
        $SshShell.Close()
        $SshShell = $null
        Exit
    }
}


# close/remove SSH session
function Remove-SshSession {
    Param ( [Tamir.SharpSsh.SshShell] $SshShell )

    $SshShell.WriteLine( "exit" )
    Start-Sleep -m $config.SleepTime
    if ( $SshShell.ShellOpened ) {
        Write-Warning "Shell didn't exit cleanly, closing anyway."
    }
    $SshShell.Close()
    $SshShell = $null

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Remove-SshSession] - SSH session removed!"
}


# Remotely compiles a jBASE Basic source file residing on a jBASE/T24 server
function Invoke-JbcCompiler {
    Param ( [String] $sourcefile )

    if ($sourcefile.Length -eq 0) { return }

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Invoke-JbcCompiler] - Compiling: $sourcefile ..."

    # create new SSH connection
    $SshShell = New-SshSession $config.Host $config.Port $config.User $config.Password

    # capture and ignore initial output after the connection was established
    $SshShell.WriteLine("")
    $output = $SshShell.Expect()

    # required when called from ue/uestudio
    $sourcefilename = $sourcefile
    if ($sourcefile.Contains("/")) {
        $sourcefilename = $sourcefile.Split("/")[-1]
    }

    if (-not $config.UseEBCOMPILE) {
        # call to jBASE commands BASIC and CATALOG
        $SshShell.WriteLine( "BASIC -v -w3 -I{0} {1} {2}" -f ($config.IncludeSrcDir, $config.RemoteSrcDir, $sourcefilename) )
        Start-Sleep -m $config.SleepTime
        $output = $SshShell.Expect()
        $output.Replace("`b", '')
        if (-not $output.Contains("** Unable to compile")) {
            $SshShell.WriteLine( "CATALOG {0} {1}" -f ($config.RemoteSrcDir, $sourcefile) )
            Start-Sleep -m $config.SleepTime
            $SshShell.Expect().Replace("`b", '')
        }
    } else {
        # call to T24 EB.COMPILE command
        $SshShell.WriteLine( "EB.COMPILE {0} {1}" -f ($config.RemoteSrcDir, $sourcefilename) )
        Start-Sleep -m $config.SleepTime
        $SshShell.Expect().Replace("`b", '')
    }

    # remove SSH connection
    Remove-SshSession $SshShell
}


# Remotely executes a jBASE Basic subroutine residing on a jBASE/T24 server
# It runs on the server a PROGRAM which in turn calls the passing
# subroutine's name and parameters. Up to ten parameters are allowed.
function Start-SubRoutine {
    Param ( [String] $subroutineName, [string[]] $subroutineArgs )

    if ($subroutineName.Length -eq 0) { return }

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-SubRoutine] - Calling subroutine: $subroutineName"
    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-SubRoutine] - Subroutine args: $subroutineArgs"

    # create new SSH connection
    $SshShell = New-SshSession $config.Host $config.Port $config.User $config.Password

    # capture and ignore initial output after the connection was established
    $SshShell.WriteLine("")
    $output = $SshShell.Expect()

    # required when called from ue/uestudio
    $subname = $subroutineName
    if ($subroutineName.Contains("/")) {
        $subname = $subroutineName.Split("/")[-1]
    }

    # execute the program responsible for calling the subroutine
    $SshShell.WriteLine( "{0} {1} {2}" -f ($config.ProgName, $subname, ($subroutineArgs -join " ")) )
    Start-Sleep -m $config.SleepTime
    $output = $SshShell.Expect().Replace("`b", '')
    $output = $output -Replace "(?m)^Calling Subroutine.*?\r\n", ""
    $output.Trim()

    # remove SSH connection
    Remove-SshSession $SshShell
}


# Remotely executes a jBASE non-interactive command at the jShell prompt.
# If the passing command asks for input then the script will halt execution.
function Invoke-JshCommand {
    Param ( [String] $cmd )

    if ($cmd.Length -eq 0) { return }

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Invoke-JshCommand] - Command: $cmd"

    # create new SSH connection
    $SshShell = New-SshSession $config.Host $config.Port $config.User $config.Password

    # capture and ignore initial output after the connection was established
    $SshShell.WriteLine("")
    $output = $SshShell.Expect()

    # send jBASE jShell command
    $SshShell.WriteLine( $cmd )
    Start-Sleep -m $config.SleepTime

    # jBASE command execution output
    $SshShell.Expect().Replace("`b", '')

    # remove SSH connection
    Remove-SshSession $SshShell
}


#   Evaluates jBASE Basic expressions / statements
#
# - Takes as parameter a string buffer containing the jBC expresions/statements.
# - Reads the jBC template file and merges it with the jBC code string.
# - The resulting string containing the source is encoded to Base64.
# - Remotely runs a program which in turn calls the subroutine that will
#   dinamically assign a subroutine name to the source code sent, then will
#   create the source file on the remote jBASE/T24 server and finally
#   will invoke the compilation of the resulting source file.
# - The subroutine responsible for the creation and compilation of the
#   source file will output to the screen the subroutine name dinamically
#   obtained.
# - A simple screen scrapping procedure will get the name of the
#   subroutine to be called.
function Start-JbcCodeEval {
    Param ( [String] $sourceString )

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-JbcCodeEval] - merge, encode to BASE64 and build subroutine"

    $jbcEvalCode = $sourceString
    $jbcTemplate = ( gc (Join-Path $scriptDir "jbcTemplate.b") -ErrorAction Stop | Out-String )
    $jbcTemplate = $jbcTemplate.Replace("NULL", $jbcEvalCode)

    # create new SSH connection
    $SshShell = New-SshSession $config.Host $config.Port $config.User $config.Password

    # create jBC source string with contents encoded then calls the subroutine
    # that will create and compile the source file on the server
    $bytes = $config.Encoding.GetBytes($jbcTemplate)
    $encodedStr = [System.Convert]::ToBase64String($bytes)
    $SshShell.WriteLine( "{0} {1} {2} {3}" -f ($config.ProgName, $config.SubName, $config.RemoteEvalDir, $encodedStr) )
    Start-Sleep -m $config.SleepTime
    $output = $SshShell.Expect("Calling Subroutine \.\.\.")

    $SshShell.WriteLine("")
    Start-Sleep -m $config.SleepTime
    $output = $SshShell.Expect().Replace("`b", '')

    if ($output.Contains("** Unable to compile source")) {
        $output.Replace($config.Prompt, '').Trim()

        # remove SSH connection
        Remove-SshSession $SshShell
    } else {
        $dynsubname = [string]::Empty
        if ($output -match 'CodeSnippetUserNo: (\S+\d+)') {
            $dynsubname = $matches[1]
            Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-JbcCodeEval] - Dynamically created subroutine: $dynsubname"

            # remove SSH connection
            Remove-SshSession $SshShell

            # Call subroutine
            Start-SubRoutine $dynsubname
        } else {
            # remove Telnet connection
            Remove-TelnetSession $telnet
        }
    }
}


#   Evaluates jBASE Basic expressions / statements
# - Takes as parameter the name of the file containing the jBC expresions.
function Start-JbcSourceFileEval {
    Param ( [String] $sourceFilePath )

    if ($sourceFilePath.Length -eq 0) { return }

    $fileInfo = Get-item "$sourceFilePath" -ErrorAction Stop
    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-JbcSourceFileEval] - Source: $($fileInfo.Name)"

    $jbcCode = ( gc $fileInfo.FullName -ErrorAction Stop | Out-String )
    if ($jbcCode.Length -gt 0) {
        Start-JbcCodeEval $jbcCode
    } else {
        Write-Host "No jBC code snippet provided for evaluation"
    }
}


#   Evaluates jBASE Basic expressions / statements
# - Takes as parameter a string encoded to Base64 containing the jBC expresions.
function Start-JbcSourceStringEval {
    Param ( [String] $sourceString )

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-JbcSourceStringEval] - ..."

    $bytes   = [System.Convert]::FromBase64String($sourceString)
    $jbcCode = $config.Encoding.GetString($bytes)
    if ($jbcCode.Length -gt 0) {
        Start-JbcCodeEval $jbcCode
    } else {
        Write-Host "No jBC code snippet provided for evaluation"
    }
}


# Upload a jBC source file to a remote jBASE/T24 server via SFTP
function Start-Upload {
    Param ( [String] $sourceFilePath )

    if ($sourceFilePath.Length -eq 0) { return }

    $fileInfo = Get-item $sourceFilePath -ErrorAction Stop
    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - SFTP Uploading file: $($fileInfo.FullName)"
    $sftp = new-object Tamir.SharpSsh.Sftp $config.Host, $config.User, $config.Password
    Try {
        $sftp.Connect()
        Start-Sleep -m 100
        Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - $($sftp.ServerVersion)"
        Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - $($sftp.ClientVersion)"
        $sftp.Put($fileInfo.FullName, "$($config.RemoteSrcDir)/$($fileInfo.Name)")
        $msg = '"{0}" has been successfully uploaded to "{1}" on "{2}"'
        $msg = $msg -f $fileInfo.Name, $config.RemoteSrcDir, $($config.Host)
        Write-Host $msg
    } Catch [System.Exception] {
        Write-Host $_.Exception.GetType().FullName
        Write-Host $_.Exception.Message
    } Finally {
        if (-not ($sftp -eq $null) -and $sftp.Connected) {
            Start-Sleep -m 100
            $sftp.Close()
        }
    }
}
