<#
    jBASE Windows Telnet Actions Module

#>


# remove telnet terminal emulation characters
function Remove-TerminalEmulationChars {
    Param ( [string] $TerminalText, [switch] $Extract )

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Remove-TerminalEmulationChars] - ..."
    $text = ""
    if ($Extract) {
        $pattern = '(?ms)(' + $config.Prompt + ')(.*)'
        if ($TerminalText -match $pattern) {
            $text = $matches[0]
        }
    } else {
        $text = $TerminalText
    }
    $text = $text -replace "\e\[H", [System.Environment]::NewLine
    $text = $text -replace "[\b]*", ""
    $text = $text -replace "\e\[[0-9;?]*[^0-9;]", ""

    return $text
}


# reads telnet's output, optionally removing terminal emulation characters
function Get-TelnetLoginOutput {
    Param ( [Hashtable] $telnet, [switch] $RemoveEmulChars, [switch] $Extract )

    $BuffSize = $telnet.Client.ReceiveBufferSize
    $Buffer = New-Object System.Byte[] $BuffSize
    $output = [string]::Empty
    While ($telnet.Stream.DataAvailable) {
        $count = $telnet.Stream.Read($Buffer, 0, $BuffSize)
        $output += ($config.Encoding.GetString($Buffer, 0, $count))
    }
    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Get-TelnetLoginOutput] - Raw output begin ------`n$output`nRaw output end ------"

    if ($RemoveEmulChars) {
        return (Remove-TerminalEmulationChars $output -Extract:$Extract)
    } else {
        return $output
    }
}


# reads telnet's output, optionally removing terminal emulation characters
function Get-TelnetOutput {
    Param ( [Hashtable] $telnet, [switch] $RemoveEmulChars, [switch] $Extract )

    $BuffSize = $telnet.Client.ReceiveBufferSize
    $Buffer = New-Object System.Byte[] $BuffSize
    $output = [string]::Empty

    $sb = New-Object System.Text.StringBuilder $BuffSize
    $Buffer = New-Object System.Byte[] $BuffSize

    # pattern should match otherwise stream read operations will wait
    # until $config.Timeout has expired
    $pattern = $config.Prompt + "\e\[K\e\[J$|jBASE debugger->$|\*\* Error "
    $regex = New-Object Regex($pattern)
    Try {
        do {
            Start-Sleep -m $config.SleepTime
            $count = $telnet.Stream.Read($Buffer, 0, $BuffSize)
            $sb.Append( $config.Encoding.GetString($Buffer, 0, $count) ) | Out-Null
            $match = $regex.Matches($sb.ToString())
        } until ( $match[0].Success )
    } Catch [System.IO.IOException] {
        $sb.Append("`nTimed out waiting for reply from telnet read operation") | Out-Null
    }
    $output = $sb.ToString()
    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Get-TelnetOutput] - Raw output begin ------`n$output`nRaw output end ------"

    if ($RemoveEmulChars) {
        return (Remove-TerminalEmulationChars $output -Extract:$Extract)
    } else {
        return $output
    }
}


# create Windows Telnet session
function New-TelnetSession {
    Param (
        [string] $HostName,
        [int] $PortNum,
        [string] $UserName,
        [string] $UserPwd
    )

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [New-TelnetSession] - Connecting via Telnet to jBASE server on $($HostName):$($PortNum)"

    Try {
        [Hashtable] $telnet = @{}
        $telnet.Client = New-Object System.Net.Sockets.TcpClient($HostName, $PortNum)
        $telnet.Stream = $telnet.Client.GetStream()
        if ($config.Timeout -eq 0) {
            $telnet.Stream.ReadTimeout = [System.Threading.Timeout]::Infinite
        } else {
            $telnet.Stream.ReadTimeout = $config.Timeout
        }

        $LoginCredentials = $UserName, $UserPwd
        ForEach ($Item in $LoginCredentials) {
            $msg = $config.Encoding.GetBytes($Item + [System.Environment]::NewLine)
            $telnet.Stream.Write($msg, 0, $msg.Length)
            Start-Sleep -m $config.SleepTime
        }

        # validate login
        $output = Get-TelnetLoginOutput $telnet -RemoveEmulChars
        if ($output.Contains("Logon failed")) {
            Throw [System.Exception] "Logon failed: unknown user name, password or privilege incorrect."
        }

        # send response to jShell message prompt: START GLOBUS Y/N=
        $msg = $config.Encoding.GetBytes("N" + [System.Environment]::NewLine)
        $telnet.Stream.Write($msg, 0, $msg.Length)
        Start-Sleep -m $config.SleepTime

        Write-Verbose "$(Get-Date -f $config.Dtfmt) [New-TelnetSession] - Telnet session created!"

        $telnet

    } Catch [System.Exception] {
        Write-Warning "Verfy connection properties in 'jbcConfig'"
        Write-Host $_.Exception.InnerException
        Write-Host $_.Exception.GetType().FullName
        Write-Host $_.Exception.Message
        if ( $telnet.Stream ) { $telnet.Stream.Close() }
        if ( $telnet.Client ) { $telnet.Client.Close() }
        $telnet = $null
        Exit
    }
}


# close / remove Telnet session
function Remove-TelnetSession {
    Param ( [Hashtable] $telnet )

    # send command for closing telnet session
    $msg = $config.Encoding.GetBytes("     off" + [System.Environment]::NewLine)
    $telnet.Stream.Write($msg, 0, $msg.Length)
    Start-Sleep -m $config.SleepTime

    $telnet.Stream.Close()
    $telnet.Client.Close()
    $telnet = $null

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Remove-TelnetSession] - Telnet session removed!"
}


# Remotely compiles a jBASE Basic source file residing on a jBASE/T24 server
function Invoke-JbcCompiler {
    Param ( [String] $sourcefile )

    if ($sourcefile.Length -eq 0) { return }

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Invoke-JbcCompiler] - Compiling: $sourcefile ..."

    # create new Telnet connection
    $telnet = New-TelnetSession $config.Host $config.Port $config.User $config.Password

    # required when called from ue/uestudio
    $sourcefilename = $sourcefile
    if ($sourcefile.Contains("/")) {
        $sourcefilename = $sourcefile.Split("/")[-1]
    }

    if (-not $config.UseEBCOMPILE) {
        # call to jBASE commands BASIC and CATALOG
        $cmd = "BASIC -v -w3 -I{0} {1} {2}" -f $config.IncludeSrcDir, $config.RemoteSrcDir, $sourcefilename
        $msg = $config.Encoding.GetBytes($cmd + [System.Environment]::NewLine)
        $telnet.Stream.Write($msg, 0, $msg.Length)
        Start-Sleep -m $config.SleepTime
        $output = Get-TelnetOutput $telnet -RemoveEmulChars -Extract
        if (-not $output.Contains("** Unable to compile source")) {
            $cmd = "CATALOG {0} {1}" -f $config.RemoteSrcDir, $sourcefilename
            $msg = $config.Encoding.GetBytes($cmd + [System.Environment]::NewLine)
            $telnet.Stream.Write($msg, 0, $msg.Length)
            Start-Sleep -m $config.SleepTime
            $output += Get-TelnetOutput $telnet -RemoveEmulChars
        }
    } else {
        # call to T24 EB.COMPILE command
        $cmd = "EB.COMPILE {0} {1}" -f $config.RemoteSrcDir, $sourcefilename
        $msg = $config.Encoding.GetBytes($cmd + [System.Environment]::NewLine)
        $telnet.Stream.Write($msg, 0, $msg.Length)
        Start-Sleep -m $config.SleepTime
        $output = Get-TelnetOutput $telnet -RemoveEmulChars -Extract
    }
    Write-Host $output

    # remove Telnet connection
    Remove-TelnetSession $telnet
}


# Remotely executes a jBASE Basic subroutine residing on a jBASE/T24 server
# It runs on the server a PROGRAM which in turn calls the passing
# subroutine's name and parameters. Up to ten parameters are allowed.
function Start-SubRoutine {
    Param ( [String] $subroutineName, [string[]] $subroutineArgs )

    if ($subroutineName.Length -eq 0) { return }

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-SubRoutine] - Calling subroutine: $subroutineName"
    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-SubRoutine] - Subroutine args: $subroutineArgs"

    # create new Telnet connection
    $telnet = New-TelnetSession $config.Host $config.Port $config.User $config.Password

    # required when called from ue/uestudio
    $subname = $subroutineName
    if ($subroutineName.Contains("/")) {
        $subname = $subroutineName.Split("/")[-1]
    }

    # execute the program responsible for calling the subroutine
    $cmd = "{0} {1} {2}" -f $config.ProgName, $subname, ($subroutineArgs -join " ")
    $msg = $config.Encoding.GetBytes($cmd + [System.Environment]::NewLine)
    $telnet.Stream.Write($msg, 0, $msg.Length)
    Start-Sleep -m $config.SleepTime
    $output = Get-TelnetOutput $telnet -RemoveEmulChars -Extract
    Write-Host ($output -replace "(?m)^Calling Subroutine \.{3}.*?\r\n", '')

    # remove Telnet connection
    Remove-TelnetSession $telnet
}


# Remotely executes a jBASE non-interactive command at the jShell prompt.
# If the passing command asks for input then the script will halt execution.
function Invoke-JshCommand {
    Param ( [String] $cmd )

    if ($cmd.Length -eq 0) { return }

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Invoke-JshCommand] - Command: $cmd"

    # create new Telnet connection
    $telnet = New-TelnetSession $config.Host $config.Port $config.User $config.Password

    # send jBASE jShell command
    $msg = $config.Encoding.GetBytes($cmd + [System.Environment]::NewLine)
    $telnet.Stream.Write($msg, 0, $msg.Length)
    Start-Sleep -m $config.SleepTime
    $output = Get-TelnetOutput $telnet -RemoveEmulChars -Extract
    Write-Host $output

    # remove Telnet connection
    Remove-TelnetSession $telnet
}


#   Evaluates jBASE Basic expressions / statements
#
# - Takes as parameter a string buffer containing the jBC expresions/statements.
# - Reads the jBC template file and merges it with the jBC code string.
# - The resulting string containing the source is encoded to Base64.
# - Remotely runs a program which in turn calls the subroutine that will
#   dinamically assign a subroutine name to the source code sent, then will
#   create the source file on the remote jBASE/T24 server and finally
#   will invoke the compilation of the resulting source file.
# - The subroutine responsible for the creation and compilation of the
#   source file will output to the screen the subroutine name dinamically
#   obtained.
# - A simple screen scrapping procedure will get the name of the
#   subroutine to be called.
function Start-JbcCodeEval {
    Param ( [String] $sourceString )

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-JbcCodeEval] - merge, encode to BASE64 and build subroutine"

    $jbcEvalCode = $sourceString
    $jbcTemplate = ( gc (Join-Path $scriptDir "jbcTemplate.b") -ErrorAction Stop | Out-String )
    $jbcTemplate = $jbcTemplate.Replace("NULL", $jbcEvalCode)

    # create new Telnet connection
    $telnet = New-TelnetSession $config.Host $config.Port $config.User $config.Password

    # create jBC source string with contents encoded then calls the subroutine
    # that will create and compile the source file on the server
    $bytes  = $config.Encoding.GetBytes($jbcTemplate)
    $encodedStr = [System.Convert]::ToBase64String($bytes)
    $cmd = "{0} {1} {2} {3}" -f $config.ProgName, $config.SubName, $config.RemoteEvalDir, $encodedStr
    $msg = $config.Encoding.GetBytes($cmd + [System.Environment]::NewLine)
    $telnet.Stream.Write($msg, 0, $msg.Length)
    Start-Sleep -m $config.SleepTime

    $output = Get-TelnetOutput $telnet -RemoveEmulChars
    if ($output.Contains("** Unable to compile source")) {
        if ($output -match '(?ms)(CodeSnippetUserNo: .*)') {
            $output = $matches[1]
        }
        Write-Host $output

        # remove Telnet connection
        Remove-TelnetSession $telnet
    } else {
        $dynsubname = [string]::Empty
        if ($output -match 'CodeSnippetUserNo: (\S+\d+)') {
            $dynsubname = $matches[1]
            Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-JbcCodeEval] - Dynamically created subroutine: $dynsubname"

            # remove Telnet connection
            Remove-TelnetSession $telnet

            # call subroutine
            Start-SubRoutine $dynsubname
        } else {
            # remove Telnet connection
            Remove-TelnetSession $telnet
        }
    }
}


#   Evaluates jBASE Basic expressions / statements
# - Takes as parameter the name of the file containing the jBC expresions.
function Start-JbcSourceFileEval {
    Param ( [String] $sourceFilePath )

    if ($sourceFilePath.Length -eq 0) { return }

    $fileInfo = Get-item $sourceFilePath -ErrorAction Stop
    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-JbcSourceFileEval] - Source: $($fileInfo.Name)"
    $jbcCode = ( gc $fileInfo.FullName -ErrorAction Stop | Out-String )
    if ($jbcCode.Length -gt 0) {
        Start-JbcCodeEval $jbcCode
    } else {
        Write-Host "No jBC code snippet provided for evaluation"
    }
}


#   Evaluates jBASE Basic expressions / statements
# - Takes as parameter a string encoded to Base64 containing the jBC expresions.
function Start-JbcSourceStringEval {
    Param ( [String] $sourceString )

    Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-JbcSourceStringEval] - ..."

    $bytes   = [System.Convert]::FromBase64String($sourceString)
    $jbcCode = $config.Encoding.GetString($bytes)
    if ($jbcCode.Length -gt 0) {
        Start-JbcCodeEval $jbcCode
    } else {
        Write-Host "No jBC code snippet provided for evaluation"
    }
}


# Upload a jBC source file to a remote jBASE/T24 server via FTP
function Start-Upload {
    Param ( [String] $sourceFilePath )

    if ($sourceFilePath.Length -eq 0) { return }

    Try {
        $fileInfo = Get-item $sourceFilePath -ErrorAction Stop
        Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - FTP Uploading file: $($fileInfo.FullName)"
        $uri = "ftp://{0}/{1}/{2}" -f $config.Host, $config.RemoteSrcDir, $fileInfo.Name
        # create FTP Request
        $ftp = [System.Net.FtpWebRequest]::Create($uri)
        $ftp.Method = [System.Net.WebRequestMethods+Ftp]::UploadFile
        $ftp.Credentials = new-object System.Net.NetworkCredential($config.User, $config.Password)
        $ftp.UseBinary = $true
        $ftp.UsePassive = $true
        $ftp.Proxy = $null

        $ftpresponse = $ftp.GetResponse()
        Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - BannerMessage: $($ftpresponse.BannerMessage)"
        Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - WelcomeMessage: $($ftpresponse.WelcomeMessage)"

        # read in the file to upload as a byte array
        $content = [System.IO.File]::ReadAllBytes($fileInfo.FullName)
        $ftp.ContentLength = $content.Length
        # get the request stream, and write the bytes into it
        $ftpstream = $ftp.GetRequestStream()
        $ftpstream.Write($content, 0, $content.Length)

        Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - ResponseUri: $($ftpresponse.ResponseUri)"
        Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - StatusCode: $($ftpresponse.StatusCode)"
        Write-Verbose "$(Get-Date -f $config.Dtfmt) [Start-Upload] - StatusDescription: $($ftpresponse.StatusDescription)"

        $msg = '"{0}" has been successfully uploaded to "{1}" on "{2}"'
        $msg = $msg -f $fileInfo.Name, $config.RemoteSrcDir, $($config.Host)
        Write-Host $msg
    } Catch [System.Exception] {
        Write-Host $_.Exception.GetType().FullName
        Write-Host $_.Exception.Message
    } Finally {
        if ($ftpstream) {
            $ftpstream.Close()
        }
        if ($ftpresponse) {
            $ftpresponse.Close()
        }
    }
}
